# TPK4186 - Advanced Tools for Performance Engineering
# LearningDelayCurves.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# It is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required Modules
# 2. Curves
# 3. Functions to learn
# 4. Training
# 5. Testing

# 1. Required Modules
# -------------------

import random
import math
import numpy as np
import matplotlib.pyplot as plt

# 2. Curves
# ---------

class Point:
  def __init__(self, x, y):
    self.x = x
    self.y = y

class Curve:
  def __init__(self):
    self.points = []
    self.sorted = True

  def Reset(self):
    self.points = []
    self.sorted = True

  def GetNumberOfPoints(self):
    return len(self.points)

  def GetPoint(self, index):
    if index<0 or index>=self.GetNumberOfPoints():
      return None
    return self.points[index]

  def AddPoint(self, x, y):
    self.points.append(Point(x, y))
    self.sorted = False

  def RemovePoint(self, index):
    if index<0 or index>=self.GetNumberOfPoints():
      return
    return self.points.pop(index)

  def Sort(self):
    if self.sorted:
      return
    self.points.sort(key=lambda point: point.x)
    self.sorted = True

  def InterpolateValue(self, x):
    if self.GetNumberOfPoints()<2:
      return None
    self.Sort()
    index = 0
    while True:
      point0 = self.GetPoint(index)
      point1 = self.GetPoint(index+1)
      if x<point1.x or index+2==self.GetNumberOfPoints():
        break
      index += 1
    return point0.y + (point1.y-point0.y)*(x-point0.x)/(point1.x-point0.x)

  def ExtractKaplanMeierCurve(self, numberOfPoints):
    kaplanMeierCurve = Curve()
    kaplanMeierCurve.AddPoint(0, 0)
    for point in self.points:
      kaplanMeierCurve.AddPoint(point.x, point.y)
    kaplanMeierCurve.Sort()
    for index in range(0, kaplanMeierCurve.GetNumberOfPoints()):
      point = kaplanMeierCurve.points[index]
      point.x = index/numberOfPoints
    return kaplanMeierCurve

# 3. Functions to learn
# ---------------------

def CalculateLinearDelay(lo, hi, z):
  return lo + (hi-lo)*z

def CalculateExponentialDelay(rate, z):
  # z = p(d) = 1.0 - exp(-rate*d)
  # d = - log(1 - z)/rate
  return -math.log(1.0 - z)/rate

# z = p(d) = 1 - exp(-pow(d/alpha, beta))
# d = pow(-log(1 - z), 1/beta)*alpha
def CalculateWeibullDelay(alpha, beta, z):
  return alpha*math.pow(-math.log(1.0 - z), 1/beta)

def CalculateBathCurveDelay(z):
  alpha1 = 1.0e4
  beta1 = 0.5
  pi1 = 2000
  alpha2 = 1.0e4
  beta2 = 1.0
  pi2 = 5000
  alpha3 = 1.0e4
  beta3 = 3
  delay = CalculateWeibullDelay(alpha1, beta1, z)
  if delay<=pi1:
      return delay
  z1 = z - 1.0 + math.exp(-(pi1/alpha1)**beta1)
  delay = pi1 + CalculateWeibullDelay(alpha2, beta2, z1)
  if delay<=pi2:
    return delay
  z2 = z1 - 1.0 + math.exp(-((pi2-pi1)/alpha2)**beta2)
  delay = pi2 + CalculateWeibullDelay(alpha3, beta3, z2)
  return delay
 

def CalculateDelay(z):
#  return CalculateLinearDelay(5000, 10000, z)
#  return CalculateExponentialDelay(1.0e-5, z)
#  return CalculateWeibullDelay(1.0e5, 3, z)
  return CalculateBathCurveDelay(z)


# 4. Training
# -----------

def TrainDelayCurve(numberOfPoints, missionTime):
  curve = Curve()
  for i in range(0, numberOfPoints):
    x = random.random()
    y = CalculateDelay(x)
    if y<=missionTime:
      curve.AddPoint(x, y)
  trainedCurve = curve.ExtractKaplanMeierCurve(numberOfPoints)
  return trainedCurve

sizeTrainingSet = 100
missionTime = 87600
trainedCurve = TrainDelayCurve(sizeTrainingSet, missionTime)

# 5. Testing
# ----------

def TestDelayCurve(curve, sizeTestingSet, missionTime):
  points = []
  for i in range(0, sizeTestingSet):
    probability = random.random()
    actualDelay = CalculateDelay(probability)
    if actualDelay<=missionTime:
      interpolatedDelay = curve.InterpolateValue(probability)
      points.append((probability, actualDelay, interpolatedDelay))
  return points

sizeTestingSet = 1000
points = TestDelayCurve(trainedCurve, sizeTestingSet, missionTime)
points.sort(key = lambda point: point[0])

probabilities = []
actualDelays = []
interpolatedDelays = []
for point in points:
  probabilities.append(point[0])
  actualDelays.append(point[1])
  interpolatedDelays.append(point[2])
plt.plot(actualDelays, probabilities, color="blue")
plt.plot(interpolatedDelays, probabilities, color="green")
plt.show()


