# TPK4186 - Advanced Tools for Performance Engineering
# PERT.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. Activity
# 3. Project

# 1. Required modules
# -------------------

import sys
import re
import random
import numpy

# 2. Activity
# -----------

class Activity:
  def __init__(self, code):
    self.code = code
    self.description = ""
    self.minimumDuration = 0
    self.mostLikelyDuration = 0
    self.maximumDuration = 0
    self.duration = 0
    self.predecessors = []
    self.successors = []
    self.earlyStartingDate = 0
    self.earlyCompletionDate = 0
    self.lateStartingDate = 0
    self.lateCompletionDate = 0
    self.criticalCount = 0

  def GetCode(self):
    return self.code

  def GetDescription(self):
    return self.description

  def SetDescription(self, description):
    self.description = description

  def GetMinimumDuration(self):
    return self.minimumDuration

  def SetMinimumDuration(self, duration):
    self.minimumDuration = duration

  def GetMostLikelyDuration(self):
    return self.mostLikelyDuration

  def SetMostLikelyDuration(self, duration):
    self.mostLikelyDuration = duration

  def GetMaximumDuration(self):
    return self.maximumDuration

  def SetMaximumDuration(self, duration):
    self.maximumDuration = duration

  def GetDuration(self):
    return self.duration

  def DrawDuration(self):
    self.duration = int(random.triangular(self.minimumDuration, self.maximumDuration, self.mostLikelyDuration))

  def GetPredecessors(self):
    return self.predecessors

  def GetSuccessors(self):
    return self.successors

  def AddPredecessor(self, activity):
    if activity in self.predecessors:
      return
    self.predecessors.append(activity)
    activity.successors.append(self)

  def GetEarlyStartingDate(self):
    return self.earlyStartingDate

  def SetEarlyStartingDate(self, date):
    self.earlyStartingDate = date

  def GetEarlyCompletionDate(self):
    return self.earlyCompletionDate

  def SetEarlyCompletionDate(self, date):
    self.earlyCompletionDate = date

  def GetLateStartingDate(self):
    return self.lateStartingDate

  def SetLateStartingDate(self, date):
    self.lateStartingDate = date

  def GetLateCompletionDate(self):
    return self.lateCompletionDate

  def SetLateCompletionDate(self, date):
    self.lateCompletionDate = date

  def IsCritical(self):
    return self.earlyStartingDate==self.lateStartingDate

  def GetCriticalCount(self):
    return self.criticalCount

  def SetCriticalCount(self, value):
    self.criticalCount = value

  def UpdateCriticalCount(self):
    if self.IsCritical():
      self.criticalCount += 1

# 3. Project
# -----------

class Project:
  def __init__(self):
    self.activities = []
    self.earlyCompletionDate = 0

  def GetActivities(self):
    return self.activities

  def LookForActivity(self, code):
    for activity in self.activities:
      if activity.GetCode()==code:
        return activity
    return None

  def NewActivity(self, code):
    activity = self.LookForActivity(code)
    if activity==None:
      activity = Activity(code)
      self.activities.append(activity)
    return activity

  def RecordActivity(self, code, description, duration, predecessors):
    activity = self.NewActivity(code)
    activity.SetDescription(description)
    activity.SetDuration(duration)
    for predecessor in predecessors:
      activity.AddPredecessor(predecessor)
    return activity

  def GetEarlyCompletionDate(self):
    return self.earlyCompletionDate

  def DrawActivityDurations(self):
    for activity in self.GetActivities():
      activity.DrawDuration()

  def ComputeEarlyDates(self):
    # Here we take benefit that if an activity A has been declared before an activity B
    # then B cannot precede A
    self.earlyCompletionDate = 0
    for activity in self.GetActivities():
      activity.SetEarlyStartingDate(0)
      for predecessor in activity.GetPredecessors():
        if predecessor.GetEarlyCompletionDate()>activity.GetEarlyStartingDate():
          activity.SetEarlyStartingDate(predecessor.GetEarlyCompletionDate())
      activity.SetEarlyCompletionDate(activity.GetEarlyStartingDate() + activity.GetDuration())
      if activity.GetEarlyCompletionDate()>self.earlyCompletionDate:
        self.earlyCompletionDate = activity.GetEarlyCompletionDate()

  def ComputeLateDates(self):
    activities = self.GetActivities()[:] # copy the list of activities
    activities.reverse()
    for activity in activities:
      activity.SetLateCompletionDate(self.earlyCompletionDate)
      for successor in activity.GetSuccessors():
        if successor.GetLateStartingDate()<activity.GetLateCompletionDate():
          activity.SetLateCompletionDate(successor.GetLateStartingDate())
      activity.SetLateStartingDate(activity.GetLateCompletionDate() - activity.GetDuration())

  def ExtractCriticalActivities(self):
    criticalActivities = []
    for activity in self.GetActivities():
      if activity.IsCritical():
        criticalActivities.append(activity)
    return criticalActivities

  def ResetCriticalCounts(self):
    for activity in self.GetActivities():
      activity.SetCriticalCount(0)

  def UpdateCriticalCounts(self):
    for activity in self.GetActivities():
      activity.UpdateCriticalCount()

  def PerformSimulations(self, numberOfSimulations):
    self.ResetCriticalCounts()
    projectDurations = numpy.zeros(numberOfSimulations)
    for index in range(0, numberOfSimulations):
      self.DrawActivityDurations()
      self.ComputeEarlyDates()
      projectDurations[index] = self.earlyCompletionDate
      self.ComputeLateDates()
      self.UpdateCriticalCounts()
    return projectDurations