# TPK4186 - Advanced Tools for Performance Engineering
# TSVInterface.py
# by Antoine Rauzy
# Copyrights (c) 2019-2020 NTNU
#
# This file is part of the educational material provided to NTNU students registered to the course TPK4186.
# This work is licensed under a Creative Commons Attribution-NonCommercial 4.0 International License.
# For details about this license, see https://creativecommons.org/licenses/by-nc/4.0/.

# Table of Contents
# -----------------
# 1. Required modules
# 2. TSV Printer
# 3. TSV Parser

# 1. Required modules
# -------------------

from PERT import *

# 2. TSV Printer
# --------------

class TSVPrinter:
  def __init__(self):
    self.separator = "\t"
    self.predecessorSeparator = ","

  def ExportProjectAtTSVFormat(self, project, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.PrintProjectAtTSVFormat(project, output)
    output.close()
    return 0

  def PrintProjectAtTSVFormat(self, project, output):
    output.write("Code" + self.separator)
    output.write("Description" + self.separator)
    output.write("Predecessors" + self.separator)
    output.write("Optimistic duration" + self.separator)
    output.write("Most likely duration" + self.separator)
    output.write("Pessimistic duration" + "\n")
    for activity in project.GetActivities():
      self.PrintActiviyAtTSVFormat(activity, output)
    output.flush()

  def PrintActiviyAtTSVFormat(self, activity, output):
    output.write(activity.GetCode() + self.separator)
    output.write(activity.GetDescription() + self.separator)
    first = True
    for predecessor in activity.GetPredecessors():
      if first:
        first = False
      else:
        output.write(self.predecessorSeparator)
      output.write(predecessor.GetCode())
    output.write(self.separator)
    output.write(str(activity.GetMinimumDuration()) + self.separator)
    output.write(str(activity.GetMostLikelyDuration()) + self.separator)
    output.write(str(activity.GetMaximumDuration()) + "\n")


# 3. TSV Parser
# -------------

class TSVParser:
  def __init__(self):
    self.separator = "\t"
    self.predecessorSeparator = ","
    self.errors = 0
    self.lineNumber = 0

  def ImportProjectAtTSVFormat(self, project, fileName):
    try:
      input = open(fileName, "r")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.errors = 0
    self.ParseProjectAtTSVFormat(project, input)
    input.close()
    return self.errors

  def ParseProjectAtTSVFormat(self, project, input):
    self.lineNumber = 0
    for line in input:
      self.lineNumber += 1
      if self.lineNumber==1: # Header
        continue
      self.ParseActivityAtTSVFormat(project, line)

  def ParseActivityAtTSVFormat(self, project, line):
    line.rstrip()
    tokens = line.split(self.separator)
    if len(tokens)!=6:
      self.RaiseError('Error line {0:d} : wrong number of elements\n'.format(self.lineNumber))
      return
    code = tokens[0]
    activity = project.LookForActivity(code)
    if activity!=None:
      self.RaiseError('Error line {0:d}: activity {0:s} already declared\n'.format(self.lineNumber, code))
      return
    activity = project.NewActivity(code)
    activity.SetDescription(tokens[1])
    codes = tokens[2].split(self.predecessorSeparator)
    for code in codes:
      if code=='':
        continue
      code = code.lstrip()
      code = code.rstrip()
      predecessor = project.LookForActivity(code)
      if predecessor==None:
        self.RaiseError('Error line {0:d}: undeclared activity {1:s}\n'.format(self.lineNumber, code))
      else:
        activity.AddPredecessor(predecessor)
    if self.IsInteger(tokens[3]):
     activity.SetMinimumDuration(int(tokens[3]))
    else:
      self.RaiseError('Error line {0:d}: incorrect minimum duration\n'.format(self.lineNumber))
    if self.IsInteger(tokens[4]):
      activity.SetMostLikelyDuration(int(tokens[4]))
    else:
      self.RaiseError('Error line {0:d}: incorrect most likely duration\n'.format(self.lineNumber))
    if self.IsInteger(tokens[5]):
      activity.SetMaximumDuration(int(tokens[5]))
    else:
      self.RaiseError('Error line {0:d}: incorrect minimum duration\n'.format(self.lineNumber))

  def RaiseError(self, message):
    sys.stderr.write(message)
    sys.stderr.flush()
    self.errors += 1

  def IsInteger(self, token):
    return re.match("[0-9]+", token)

  def IsFloat(self, token):
    return re.match("[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?", token)


