# TPK4186 - Advanced Tools for Performance Engineering
# Case study: Sales Analysis
# by Antoine Rauzy
# This work is licensed under a Creative Commons Attribution-ShareAlike 4.0 International License
# https://creativecommons.org/licenses/by-sa/4.0/

# Table of contents
# -----------------
# 1) Imported modules
# 2) Products
# 3) Shop
#   3.1) Products
#   3.2) Export by lines at TSV format
#   3.3) Export by lines at XLSX format
#   3.4) Export by lines at LaTeX format
#   3.5) Export by rows at TSV format
#   3.6) Export by rows at XLSX format
# 4) Main

# 1) Imported modules
# -------------------

import sys
import string
import math
import random
import calendar
import openpyxl

# 2) Products
# --------

class Product:
  def __init__(self, reference):
    self.reference = reference
    self.profitPerSale = 1
    self.monthlySales = [0 for month in range(0, 13)]


# 3) Shop
# -------

class Shop:
  def __init__(self):
    self.products = dict()

# 3.1) Products
# ~~~~~~~~~~~~~

  def NewProduct(self, category, profitPerSale):
    code = "%d" % random.uniform(0, 1000)
    code = code.zfill(3)
    reference = category + code
    product = Product(reference)
    product.profitPerSale = profitPerSale
    self.products[reference] = product
    return product

  def CreateProduct(self, category, profitPerSale, lowSale, highSale, modeSale):
    product = self.NewProduct(category, profitPerSale)
    for month in range(1, 13):
      sale = int(math.floor(random.triangular(lowSale, highSale, modeSale)))
      product.monthlySales[month] = sale
    return product

  def CreateProducts(self):
    self.CreateProduct("BVR", 1.20, 4000, 6000, 5200)
    self.CreateProduct("BVR", 0.80, 7000, 8000, 7200)
    self.CreateProduct("BVR", 0.85, 6000, 8000, 6400)
    self.CreateProduct("BKR", 1.10, 3000, 4000, 3400)
    self.CreateProduct("BKR", 0.85, 5000, 8000, 5700)
    self.CreateProduct("DRY", 0.50, 7000, 8000, 7400)
    self.CreateProduct("DRY", 0.90, 6000, 7000, 6900)
    self.CreateProduct("PRD", 0.30, 8000, 9000, 8700)
    self.CreateProduct("PRD", 0.40, 7000, 9000, 7800)
    self.CreateProduct("PRD", 0.55, 3000, 9000, 5400)

# 3.2) Export by lines at TSV format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByLinesAtTSVFormat(self, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.PrintByLinesAtTSVFormat(output)
    output.close()
    return 0

  def PrintByLinesAtTSVFormat(self, output):
    output.write("Product")
    for product in self.products.values():
      output.write("\t%s" % product.reference)
    output.write("\n")
    output.write("Profit per sale")
    for product in self.products.values():
      output.write("\t%s" % product.profitPerSale)
    output.write("\n")
    for month in range(1, 13):
      output.write(calendar.month_abbr[month])
      for product in self.products.values():
        output.write("\t%d" % product.monthlySales[month])
      output.write("\n")
    output.flush()

# 3.3) Export by lines at XLSX format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByLinesAtXLSXFormat(self, fileName):
    workbook = openpyxl.Workbook()
    worksheet = workbook.active
    self.UploadByLinesInWorksheet(worksheet)
    workbook.save(fileName)
    return 0

  def UploadByLinesInWorksheet(self, worksheet):
    worksheet['A1'] = "Product"
    count = 1
    for product in self.products.values():
      worksheet['%s1' % chr(ord('A')+count)] = product.reference
      count += 1
    worksheet['A2'] = "Profit per sale"
    count = 1
    for product in self.products.values():
      worksheet['%s2' % chr(ord('A')+count)] = product.profitPerSale
      count += 1
    for month in range(1, 13):
      monthIndex = month+2
      worksheet['A%d' % monthIndex] = calendar.month_abbr[month]
      count = 1
      for product in self.products.values():
        worksheet['%s%d' % (chr(ord('A')+count), monthIndex) ] = product.monthlySales[month]
        count += 1

# 3.4) Export by lines at LaTeX format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByLinesAtLaTeXFormat(self, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.PrintByLineAtLaTeXFormat(output)
    output.close()
    return 0

  def PrintByLineAtLaTeXFormat(self, output):
    output.write("\\begin{tabular}{|c|")
    for product in self.products.values():
      output.write("c|")
    output.write("}\\hline\n")
    output.write("Product")
    for product in self.products.values():
      output.write(" & %s" % product.reference)
    output.write("\\\\\\hline\n")
    output.write("Profit per sale")
    for product in self.products.values():
      output.write(" & %s" % product.profitPerSale)
    output.write("\\\\\\hline\n")
    for month in range(1, 13):
      output.write(calendar.month_abbr[month])
      for product in self.products.values():
        output.write(" & %d" % product.monthlySales[month])
      output.write("\\\\\\hline\n")
    output.write("\\end{tabular}")
    output.flush()

# 3.5) Export by rows at TSV format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByRowsAtTSVFormat(self, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.PrintByRowsAtLaTeXFormat(output)
    output.close()
    return 0

  def PrintByRowsAtTSVFormat(self, output):
    output.write("Product\tProfit")
    for month in range(1, 13):
      output.write("\t" + calendar.month_abbr[month])
    output.write("\n")
    for product in self.products.values():
      output.write(product.reference + "\t" + str(product.profitPerSale))
      for month in range(1, 13):
        output.write("\t%d" % product.monthlySales[month])
      output.write("\n")
    output.flush()

# 3.6) Export by rows at XLSX format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByRowsAtXLSXFormat(self, fileName):
    workbook = openpyxl.Workbook()
    worksheet = workbook.active
    self.UploadByRowsInWorksheet(worksheet)
    workbook.save(fileName)
    return 0

  def UploadByRowsInWorksheet(self, worksheet):
    worksheet['A1'] = "Product"
    worksheet['B1'] = "Profit"
    for month in range(1, 13):
      worksheet['%s1' % chr(ord('B')+month)] = calendar.month_abbr[month]
    count = 2
    for product in self.products.values():
      worksheet['A%d' % count] = product.reference
      worksheet['B%d' % count] = str(product.profitPerSale)
      for month in range(1, 13):
        worksheet['%s%d' % (chr(ord('B')+month), count)] = product.monthlySales[month]
      count += 1

# 3.7) Export by rows at LaTeX format
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

  def ExportByRowsAtLaTeXFormat(self, fileName):
    try:
      output = open(fileName, "w")
    except:
      sys.stderr.write('Unable to open file "%s"\n' % fileName)
      sys.stderr.flush()
      return 1
    self.PrintByRowsAtLaTeXFormat(output)
    output.close()
    return 0

  def PrintByRowsAtLaTeXFormat(self, output):
    output.write("\\begin{tabular}{|c|c|c|c|c|c|c|c|c|c|c|c|c|c|}\\hline\n")
    output.write("Product & Profit")
    for month in range(1, 13):
      output.write(" & " + calendar.month_abbr[month])
    output.write("\\\\\\hline\n")
    for product in self.products.values():
      output.write(product.reference + " & " + str(product.profitPerSale))
      for month in range(1, 13):
        output.write(" & %d" % product.monthlySales[month])
      output.write("\\\\\\hline\n")
    output.write("\\end{tabular}")
    output.flush()

# 4) Main
# --------

shop = Shop()
shop.CreateProducts()

shop.ExportByRowsAtTSVFormat("sales.tsv")
shop.ExportByRowsAtXLSXFormat("sales.xlsx")
shop.ExportByRowsAtLaTeXFormat("sales.tex")

